#pragma once

namespace Hyprutils::Math {
    enum eTransform {
        HYPRUTILS_TRANSFORM_NORMAL      = 0,
        HYPRUTILS_TRANSFORM_90          = 1,
        HYPRUTILS_TRANSFORM_180         = 2,
        HYPRUTILS_TRANSFORM_270         = 3,
        HYPRUTILS_TRANSFORM_FLIPPED     = 4,
        HYPRUTILS_TRANSFORM_FLIPPED_90  = 5,
        HYPRUTILS_TRANSFORM_FLIPPED_180 = 6,
        HYPRUTILS_TRANSFORM_FLIPPED_270 = 7,
    };
}
